/* YCGM CCD data -> PSD by paul69, v0.01, 2003-01-15, <e2500@narod.ru>

   Based on
   Canon PowerShot Converter
   Copyright 1997-2002 by Dave Coffin <dcoffin@shore.net>
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

class TFile
{
public:
   TFile(FILE* p=NULL) : fp(p) {}
   ~TFile() { if(fp) fclose(fp); }

   FILE* fp;
   operator FILE* () { return fp; }

   void operator = (FILE* p) { if(fp) fclose(fp); fp = p; }
   void close() { if(fp) fclose(fp); fp=NULL; }
};

   enum { Gr, Ma, Cy, Ye };
   enum { R, G, B };

typedef unsigned char byte;
typedef unsigned short word;
typedef unsigned long dword;
typedef unsigned short color_t;

class TBuff
{
public:
   TBuff() : p(NULL) {}
   TBuff(unsigned sz) : p(new color_t[sz]) {}
   ~TBuff() { delete [] p; }

   color_t* p;
   operator color_t* () { return p; }
   color_t& operator [] (int i) { return p[i]; }

   bool alloc(unsigned sz) { delete [] p; return NULL != (p = new color_t[sz]); }
   void free() { delete [] p; p = NULL; }
};


/*
   RGB to GMCY conversion matrix
   
    0.1100    0.8600    0.0800
    0.5000    0.2900    0.5100
    0.1100    0.9200    0.7500
    0.8100    0.9800    0.0800

   C is back conversion matrix
   C = inv( A'*W*A )*A'*W, W=identity matrix

*/
void get_rgb(color_t rgb[3], color_t X[4])
{
   double temp;

   static
   float C[3][4] = {
      { -0.6086f,  0.8780f, -0.6239f,  0.8599f},
      {  0.7905f, -0.6004f,  0.3002f,  0.2225f},
      { -0.5369f,  0.8864f,  0.8507f, -0.5898f}};

   temp = C[R][Gr]*X[Gr] + C[R][Ma]*X[Ma] + C[R][Cy]*X[Cy] + C[R][Ye]*X[Ye];
   if( temp < 0 ) temp = 0; if( temp > 65535 ) temp = 65535;
   rgb[R] = (color_t) temp;

   temp = C[G][Gr]*X[Gr] + C[G][Ma]*X[Ma] + C[G][Cy]*X[Cy] + C[G][Ye]*X[Ye];
   if( temp < 0 ) temp = 0; if( temp > 65535 ) temp = 65535;
   rgb[G] = (color_t) temp;

   temp = C[B][Gr]*X[Gr] + C[B][Ma]*X[Ma] + C[B][Cy]*X[Cy] + C[B][Ye]*X[Ye];
   if( temp < 0 ) temp = 0; if( temp > 65535 ) temp = 65535;
   rgb[B] = (color_t) temp;
}

dword m4(dword x)
{
   return ((x>>24)&0xFF)|((x>>8)&0xFF00)|((x&0xFF00)<<8)|((x&0xFF)<<24);
}

word m2(dword x)
{
   return (word)(((x>>8)&0xFF)|((x&0xFF)<<8));
}

/*
   Write the image to a 48-bit Photoshop file.
*/

bool write_psd(FILE *ofp, color_t* cmyg, int width, int height)
{
   struct PSD_hdr
   {
      dword signature;
      dword version;
      dword reserved;
      word channels;
      dword height;
      dword width;
      word bits;
      word mode;
      dword color_mode;
      dword image_resources;
      dword layer_mask_info;
      word compression;
   } hdr;

   memset( &hdr, 0, sizeof(hdr) );

   hdr.signature = *(dword*)"8BPS";
   hdr.version = m4(0x00010000);
   hdr.channels = m2(3);
   hdr.bits = m2(16);
   hdr.mode = m2(3);
   hdr.width = m4(width);
   hdr.height = m4(height);

   if( !fwrite(&hdr, sizeof(hdr), 1, ofp) )
      return false;

   unsigned sz = width * height;
   unsigned size = sz * 3;

   TBuff buffer( size );
   color_t* R = &buffer[0];
   color_t* G = &buffer[sz];
   color_t* B = &buffer[sz+sz];
   color_t* im = cmyg;

   int i=0;
   for(int y = 0; y<height; ++y)
   for(int x = 0; x<width; ++x, im+=4, ++i)
   {
      color_t rgb[3];
      get_rgb(rgb, im);

      R[i] = m2(rgb[0]);
      G[i] = m2(rgb[1]);
      B[i] = m2(rgb[2]);
   }

   if( !fwrite(buffer, 2, size, ofp) )
      return false;
      
   return true;
}


struct TCCDField
{
   unsigned offset, rows;
};

struct TCCDParam
{
   unsigned raw_width;
   unsigned raw_height;
   unsigned file_size;

   char cfa_pattern[4];
   
   TCCDField field[2];
};


bool read_ccd(FILE* in, color_t ccd[], TCCDParam& ccd_pars)
{
   unsigned ccd_size = ccd_pars.raw_width * ccd_pars.raw_height;

   for(int f=0; f<2; ++f)
   {
      if( ccd_pars.field[f].rows == 0 ) continue;

      fseek( in, ccd_pars.field[f].offset, SEEK_SET );

      unsigned raw_width = ccd_pars.raw_width;
      unsigned line = f * raw_width;
      unsigned delta = 2 * raw_width;
      unsigned field_size = ccd_pars.field[f].rows * raw_width*12/8;

      unsigned i, x;

      for(x=i=0; i<field_size; i += 3)
      {
         byte bitbuf[3];
         if( !fread( bitbuf, 3, 1, in) )
            return false;

         if( x == raw_width )
         {
            line += delta;
            x = 0;

            // check space for row
            if( line + raw_width > ccd_size )
               return false;
         }

         ccd[line + x] = (color_t)(((bitbuf[0]<<4)|((bitbuf[1]>>4)&15)));
         ++x;

         ccd[line + x] = (color_t)((((bitbuf[1]&15)<<8)|(bitbuf[2])));
         ++x;
      }
   }
   return true;
}
////////////////////////////////////////////////////////////////////////////////
/*
   simple interpolation

   Y  C  Y   Y   Y  C
   G <M> G     x    |  G - G  M
   Y  C  Y   Y   Y  C

   average nearest pixels: two vertical, two horizontal, and four diagonal
*/
void near_border(color_t* in, int x, int y, int width, int height, int aver[3])
{
   int cnt[3];

   aver[0]=aver[1]=aver[2]=
   cnt[0]=cnt[1]=cnt[2]=0;

   // near border
   if( x > 0 )
   {
      aver[0] += in[-1];
      ++cnt[0];

      if( y > 0 )
      {
         aver[2] += in[-width-1];
         ++cnt[2];
      }
      if( y < height-1 )
      {
         aver[2] += in[width-1];
         ++cnt[2];
      }
   }
   if( x < width-1 )
   {
      aver[0] += in[1];
      ++cnt[0];

      if( y > 0 )
      {
         aver[2] += in[-width+1];
         ++cnt[2];
      }
      if( y < height-1 )
      {
         aver[2] += in[width+1];
         ++cnt[2];
      }
   }
   if( y > 0 )
   {
      aver[1] += in[-width];
      ++cnt[1];
   }
   if( y < height-1 )
   {
      aver[1] += in[width];
      ++cnt[1];
   }

   if( cnt[0]) aver[0] /= cnt[0];
   if( cnt[1]) aver[1] /= cnt[1];
   if( cnt[2]) aver[2] /= cnt[2];
}

bool simple_interpolation(color_t ccd[], color_t image[], TCCDParam& ccd_pars)
{
   int cfa[4]; // index of color filter

   for(int i=0; i<4; ++i)
   {
      switch( ccd_pars.cfa_pattern[i] )
      {
      case 'G': cfa[i] = Gr; break;
      case 'M': cfa[i] = Ma; break;
      case 'C': cfa[i] = Cy; break;
      case 'Y': cfa[i] = Ye; break;
      default: return false;
      }
   }

   int height = ccd_pars.raw_height;
   int height1 = height-1;
   int width = ccd_pars.raw_width;
   int width1 = width-1;

   color_t* in = ccd;
   color_t* out = image;

   for(int y=0; y<height; ++y)
   {
      for(int x=0; x<width; ++x, ++in, out += 4)
      {
         int aver[3];

         if( x == 0 || y == 0 || x == width1 || y == height1 )
         {
            near_border(in, x, y, width, height, aver);
         }
         else
         {
            aver[0] = (in[-1] + in[1])>>1;
            aver[1] = (in[-width] + in[width])>>1;
            aver[2] = (in[-width1] + in[-1-width]+in[1+width] + in[width1])>>2;
         }

         // index in CFA
         int i = (x&1) | ((y&1)<<1);
         out[ cfa[i] ] = in[0];
         out[ cfa[i^1] ] = (color_t)aver[0];
         out[ cfa[i^2] ] = (color_t)aver[1];
         out[ cfa[i^3] ] = (color_t)aver[2];
      }
   }
   return true;
}

////////////////////////////////////////////////////////////////////////////////

bool no_interpolation(color_t ccd[], color_t image[], TCCDParam& ccd_pars)
{
   int cfa[4]; // index of color filter

   for(int i=0; i<4; ++i)
   {
      switch( ccd_pars.cfa_pattern[i] )
      {
      case 'G': cfa[i] = Gr; break;
      case 'M': cfa[i] = Ma; break;
      case 'C': cfa[i] = Cy; break;
      case 'Y': cfa[i] = Ye; break;
      default: return false;
      }
   }

   int height = ccd_pars.raw_height;
   int width = ccd_pars.raw_width;

   color_t* in = ccd;
   color_t* out = image;

   for(int y=0; y<height; ++y)
   {
      for(int x=0; x<width; ++x, ++in, out += 4)
      {
         // index in CFA
         int i = (x&1) | ((y&1)<<1);
         out[ cfa[i] ] = in[0];
         out[ cfa[i^1] ] = (color_t)0;
         out[ cfa[i^2] ] = (color_t)0;
         out[ cfa[i^3] ] = (color_t)0;
      }
   }
   return true;
}

////////////////////////////////////////////////////////////////////////////////

TCCDParam default_pars[] = {
   {1616, 1204, 0x2CE000, "YCGM", {{ 0, 602 }, {0x167000, 602}}  }, // E2500
   {2288, 1708, 0x598000, "YCGM", {{ 0, 854 }, {0x2CC000, 854}}  }, // E4500
};
const int n_default_pars = sizeof(default_pars) / sizeof(default_pars[0]);

typedef bool (*interpolate_proc)(color_t[], color_t[], TCCDParam& );

void usage()
{
   printf("args: [options] input_file [output_file]\n");
   printf("options: frame pars (for hex use 0x notation)\n");
   printf("   -w:1616 - raw width\n");
   printf("   -h:1204 - raw height\n");
   printf("   -c:YCGM - CFA pattern\n");
   printf("   -b0:0 - begin of field 0\n");
   printf("   -n0:602 - lines in field 0\n");
   printf("   -b1:167000 - begin of field 1\n");
   printf("   -n1:602 - lines in field 1\n");

   printf("options: processing\n");
   printf("   -v - verbose\n");
   printf("   -i:0 - interpolation mode (0 - is default)\n");
   printf("      0 - no interpolation\n");
   printf("      1 - simple interpolation\n");
}

bool main(int n, char* a[])
{
   TCCDParam user_pars;
   memset( &user_pars, ~0, sizeof(user_pars));

   interpolate_proc interpolate_colors = no_interpolation;
   char* ifn = NULL;
   char* ofn = NULL;
   bool verbose = false;
   char ofn_buf[_MAX_PATH];

   int i;
   for(i=1; i<n; ++i)
   {
      if( a[i][0] == '-' )
      {
         switch( a[i][1] )
         {
         case 'c':
            memcpy(user_pars.cfa_pattern, &a[i][3], 4);
            break;

         case 'w':
            sscanf(&a[i][3], "%i", &user_pars.raw_width);
            break;

         case 'h':
            sscanf(&a[i][3], "%i", &user_pars.raw_height);
            break;

         case 'b': {
            int f = atoi(&a[i][2]); if( f >= 2 ) goto invalid;
            sscanf(&a[i][4], "%i", &user_pars.field[f].offset);
            } break;

         case 'n': {
            int f = atoi(&a[i][2]); if( f >= 2 ) goto invalid;
            sscanf(&a[i][4], "%i", &user_pars.field[f].rows);
            } break;

         case 'i': {
            int mode = atoi(&a[i][3]);
            if( mode == 0 ) interpolate_colors = no_interpolation;
            if( mode == 1 ) interpolate_colors = simple_interpolation;
            } break;

         case 'v': verbose = true; break;

         case '?':
            usage();
            return false;

         default:
            printf("unknown switch: %s\n", a[i]);
            break;

         invalid:
            printf("invalid switch: %s\n", a[i]);
            break;
         }
      }
      else
      {
         if( !ifn )
            ifn = a[i];
         else
            ofn = a[i];
      }
   }

   if( !ifn )
   {
      printf("error: input file is missing\n");
      return false;
   }
   if( !ofn )
   {
      char* fileName = ifn;
      char* ifn_slash = strrchr(fileName, '/');
      if( ifn_slash )
         fileName = ifn_slash;

      char* ifn_bslash = strrchr(fileName, '\\');
      if( ifn_bslash )
         fileName = ifn_bslash;

      char* ifn_ext = strrchr(fileName, '.');

      if( !ifn_ext )
         ifn_ext = ifn + strlen(ifn);

      int len = ifn_ext - ifn;

      ofn = ofn_buf;
      memcpy(ofn, ifn, len);
      strcpy(ofn+len, ".PSD");
   }

   TFile in = fopen( ifn, "rb");
   if( !in ) { printf("error: \"%s\" - can not open\n", ifn); return false; }

   fseek( in, 0, SEEK_END );
   unsigned file_size = ftell( in );

   if( verbose )
      printf("input: %s, size=%u\n", ifn, file_size);

   for(i=0; i<n_default_pars; ++i)
   {
      if( file_size == default_pars[i].file_size )
         break;
   }

   TCCDParam ccd_pars;

   if( i < n_default_pars )
   {
      if( verbose )
         printf("found default parameters for size=%u\n", file_size);

      ccd_pars = default_pars[i];
      if( user_pars.raw_width != ~0 ) ccd_pars.raw_width = user_pars.raw_width;
      if( user_pars.raw_height != ~0 ) ccd_pars.raw_height = user_pars.raw_height;
      if( user_pars.cfa_pattern[0] != -1 ) memcpy( ccd_pars.cfa_pattern, user_pars.cfa_pattern, 4);
      if( user_pars.field[0].offset != ~0 ) ccd_pars.field[0].offset = user_pars.field[0].offset;
      if( user_pars.field[0].rows != ~0 ) ccd_pars.field[0].rows = user_pars.field[0].rows;
      if( user_pars.field[1].offset != ~0 ) ccd_pars.field[1].offset = user_pars.field[1].offset;
      if( user_pars.field[1].rows != ~0 ) ccd_pars.field[1].rows = user_pars.field[1].rows;
   }
   else
   {
      if( verbose )
         printf("not found default parameters for size=%u\n", file_size);

      if( user_pars.raw_width == ~0 ) { printf("error: unknown raw width\n"); return false; }
      if( user_pars.raw_height == ~0 ) { printf("error: unknown raw height\n"); return false; }
      if( user_pars.cfa_pattern[0] == -1 ) { printf("error: unknown cfa_pattern\n"); return false; }
      if( user_pars.field[0].offset == ~0 ) { printf("error: unknown field 0 offset\n"); return false; }
      if( user_pars.field[0].rows == ~0 ) { printf("error: unknown field 0 rows\n"); return false; }
      if( user_pars.field[1].offset == ~0 ) { printf("error: unknown field 1 offset\n"); return false; }
      if( user_pars.field[1].rows == ~0 ) { printf("error: unknown field 1 rows\n"); return false; }

      ccd_pars = user_pars;
   }

   if( verbose )
   {
      printf("current parameters:\n");
      printf("raw width: %u\n", ccd_pars.raw_width);
      printf("raw height: %u\n", ccd_pars.raw_height);
      printf("cfa pattern: %4.4s\n", ccd_pars.cfa_pattern);
      printf("field 0: offset: 0x%08X rows: %u\n", ccd_pars.field[0].offset, ccd_pars.field[0].rows );
      printf("field 1: offset: 0x%08X rows: %u\n", ccd_pars.field[1].offset, ccd_pars.field[1].rows );
   }

   TBuff ccd(ccd_pars.raw_height*ccd_pars.raw_width);

   if( !ccd )
   {
      printf("error: memory allocation fault (%u)\n", ccd_pars.raw_height*ccd_pars.raw_width*sizeof(color_t));
      return false;
   }

   if( verbose )
      printf("read ccd data\n");

   if( !read_ccd( in, ccd, ccd_pars) )
   {
      printf("error: can not read ccd data\n");
      return false;
   }

   in.close();

   TBuff cmyg(ccd_pars.raw_height*ccd_pars.raw_width*4);

   if( !cmyg )
   {
      printf("error: memory allocation fault (%u)\n", ccd_pars.raw_height*ccd_pars.raw_width*4*sizeof(color_t));
      return false;
   }

   if( verbose )
      printf("interpolate colors\n");

   interpolate_colors(ccd, cmyg, ccd_pars);
   ccd.free();

   TFile out = fopen( ofn, "wb");
   if( !out ) { printf("error: \"%s\" - can not open\n", ofn); return false; }

   if( verbose )
      printf("write picture to: %s\n", ofn);

   if( !write_psd(out, cmyg, ccd_pars.raw_width, ccd_pars.raw_height) )
      { printf("error: can not write PSD\n"); return false; }

   if( verbose )
      printf("OK\n");

   return true;
}

